local SML = AceLibrary("SharedMedia-1.0")
local Threat = AceLibrary("Threat-1.0")
local dewdrop = AceLibrary("Dewdrop-2.0")

local function ConvertToK(n)
--	do return n, "" end
	if n > 1000 then
		return n / 1000, "k"
	end
	return n, ""
end

local function animateBar()
	this.animatedTime = this.animatedTime + arg1
	local pct = min(1, this.animatedTime / this.totalAnimateTime)
	local delta = (this.destinationWidth - this.startWidth) * pct
	this.tex:SetWidth(this.startWidth + delta)
	if not Omen.ActiveSkin.StretchBarTextures then
		local texPct = (this.startWidth + delta) / this.totalWidth
		this.tex:SetTexCoord(0, texPct, 0, 1)
	end
	if pct == 1 then
		this:SetScript("OnUpdate", nil)
	end
end

local function animatePercent(self, pct, dPct) 
	local w = self:GetParent():GetWidth() - 10
	self.labels.threatpercent:SetText(dPct)
	local width = w * pct
	if Omen.db.profile.AnimateBars then
		self:SetScript("OnUpdate", animateBar)
		self.totalWidth = w
		self.startWidth = self.tex:GetWidth()
		self.destinationWidth = width
		self.totalAnimateTime = 0.25
		self.animatedTime = 0
	else
		self.tex:SetWidth(width)
		if not Omen.ActiveSkin.StretchBarTextures then
			self.tex:SetTexCoord(0, pct, 0, 1)
		end
	end	
end

local function getOffset(self, data)
	local bw = self.Anchor:GetWidth() - 8
	local amt = data[1] * bw
	if data[3] == "RIGHT" then
		amt = amt * -1
	end
	return amt
end

local lookup_table = {}
local function deepcopy(object)
	for k, v in pairs(lookup_table) do
		lookup_table[k] = v
	end
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end


local default_color = {r=0,g=1,b=1,a=1}

function Omen:LoadSkin(name)
	-- self:Print("Loading skin %q", name)
	local loadedSkin = self.db.profile.Skins[name]
	local currentSettings = self.db.profile.CurrentSkinSettings
	for k, v in pairs(currentSettings) do
		currentSettings[k] = nil
	end
	_G.collectgarbage('collect') -- get rid of the copies of tables that we aren't using anymore
	for k, v in pairs(self.SkinDefaults) do
		currentSettings[k] = deepcopy((loadedSkin and loadedSkin[k]) or v)
	end	
	self.ActiveSkin = currentSettings
end

function Omen:SaveCurrentSkinAs(name)
	local skin = self.db.profile.Skins[name]
	if not skin then
		self.db.profile.Skins[name] = {}
		skin = self.db.profile.Skins[name]
	end
	for k, v in pairs(self.ActiveSkin) do
		skin[k] = v
	end
	self:UpdateSkinList()
end
------------------------------
-- Skin updates and stuff 
------------------------------

function Omen:ApplySkin()
	local skin = self.ActiveSkin
	if not skin then self:Print(debugstack()) end
	if not skin.Width then 
		self:PrintLiteral(skin)
		self:Print(debugstack())
	end
	
	-- Set the frame width
	self.Anchor:SetWidth(skin.Width)

	-- Set the frame background color
	local f = skin.BackdropColor
	self.Anchor:SetBackdropColor(f.r, f.g, f.b, f.a)

	-- Set the frame border color
	local f = skin.BorderColor
	self.Anchor:SetBackdropBorderColor(f.r, f.g, f.b, f.a)
	
	-- Set column positions
	self:UpdateLayout()
	
	-- Set bar heights, colors, textures, fonts
	for k, v in pairs(self.bars) do
		self:UpdateBarDisplay(k)
	end

	-- Set column visibility
	for k, v in pairs(self.Anchor.columns) do
		self:ToggleColumn(k, skin.ColumnVisibility[k])

		local font, size = v:GetFont()
		local cFontSettings = skin.CustomBarFonts["_COLUMNS_"]
		if cFontSettings and cFontSettings[4] then
			font = SML:Fetch("font", cFontSettings[4])
		end
		size = (cFontSettings and cFontSettings[1]) or size
		v:SetFont(font, size)		
		if cFontSettings then
			local c = cFontSettings[3]
			v:SetVertexColor(c.r, c.g, c.b, 1)
		else
			v:SetVertexColor(1,1,1,1)
		end
	end
	
	self:ArrangeBars()
end

function Omen:UpdateBarDisplay(name)
	local skin = self.ActiveSkin
	
	local v = self.bars[name]
	if not v then return end

	v.tex:SetTexture(SML:Fetch("statusbar", skin.CustomBarTextures[name] or skin.BarTexture or "Blizzard"))
	
	local c
	
	if v.unitid then
		c = skin.CustomColors[name] or RAID_CLASS_COLORS[select(2, UnitClass(v.unitid))] or default_color
	else
		c = skin.CustomColors[name] or default_color
	end
	v.tex:SetVertexColor(c.r, c.g, c.b)
	
	local h = skin.CustomBarHeights[name] or skin.BarHeight
	v:SetHeight(h)
	
	-- Need to set fonts in here too
	local cFontSettings = skin.CustomBarFonts[name] or skin.CustomBarFonts["_BARS_"]
	for k, l in pairs(v.labels) do
		local font, size = l:GetFont()
		if cFontSettings and cFontSettings[4] then
			font = SML:Fetch("font", cFontSettings[4])
		end
		if cFontSettings and cFontSettings[3] then
			l:SetVertexColor(cFontSettings[3].r, cFontSettings[3].g, cFontSettings[3].b, 1)
		end
		
		size = (cFontSettings and cFontSettings[1]) or size
		l:SetFont(font, size)
	end
end

-- Updates column and label positions
function Omen:UpdateLayout()
	for name, column in pairs(self.Anchor.columns) do
		local posData = self.ActiveSkin.ColumnOffsets[name]
		if not posData then
			self:PrintLiteral(self.ActiveSkin)
		end
		column:ClearAllPoints()
		column:SetPoint("TOP" .. posData[2], self.Anchor, "TOP" .. posData[3], getOffset(self, posData), self.titleOffset * -1 - 7)	

		for k, bar in pairs(self.bars) do
			local label = bar.labels[name]
			label:ClearAllPoints()
			label:SetPoint(posData[2], bar, posData[3], getOffset(self, posData), 0)	
		end
	end
end

function Omen:ToggleColumn(name, value)
	self.ActiveSkin.ColumnVisibility[name] = value
	if value then
		self.Anchor.columns[name]:Show()
		for k,v in pairs(self.bars) do
			v.labels[name]:Show()
		end
	else
		self.Anchor.columns[name]:Hide()
		for k,v in pairs(self.bars) do
			v.labels[name]:Hide()
		end
	end
end

------------------------------
-- Creation methods
------------------------------

function Omen:CreateBar()
	local f = CreateFrame("Button", "OmenBar" .. #self.bars, self.Anchor)
	f:SetHighlightTexture("Interface\\Buttons\\UI-Listbox-Highlight")
	-- f:GetHighlightTexture():SetBlendMode("ADD")
	
	f:RegisterForDrag("LeftButton")
	f:RegisterForClicks("RightButtonDown")
	f:SetScript("OnClick", function()
		if arg1 == "RightButton" then
			Omen.lastBarClicked = this
			dewdrop:Open(this, "children", Omen.barMenu, 'cursorX', true, 'cursorY', true)
		end
	end)
	
	f:SetScript("OnDragStart", function()
		if not this.unitid or this.unitid == "player" then return end
		local bar = OmenThreatBar:new()
		bar:CreateBar(this.unitname, this.unitid, SML:Fetch("statusbar", Omen.ActiveSkin.BarTexture))
		bar:Appear()

		bar.Bar:ClearAllPoints()
		local x, y = GetCursorPosition()
		x = x / UIParent:GetScale() - 10
		y = y / UIParent:GetScale() - 10
		bar.Bar:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)
		bar.Bar:SetFrameStrata("HIGH")
		bar.Bar:StartMoving()
		bar.Bar:SetFrameLevel(5)
		
		bar.Bar:SetScript("OnReceiveDrag", function()
			this:StopMovingOrSizing()
		end)

		local level = bar.Bar:GetFrameLevel()
	end)
	
	f.tex = f:CreateTexture()
	f.tex:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
	f.tex:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", 0, 0)
	f.tex:SetWidth(25)
	
	f.myArrow = f:CreateTexture()
	f.myArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.myArrow:SetPoint("RIGHT", f, "LEFT", -5, 0)
	f.myArrow:SetTexCoord(0, 0.5, 0, 1)
	f.myArrow:SetWidth(6)
	f.myArrow:SetHeight(12)
	f.myArrow:Hide()
	
	f.aggroArrow = f:CreateTexture()
	f.aggroArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.aggroArrow:SetPoint("LEFT", f, "RIGHT", 5, 0)
	f.aggroArrow:SetTexCoord(0.5, 1, 0, 1)
	f.aggroArrow:SetWidth(6)
	f.aggroArrow:SetHeight(12)	
	f.aggroArrow:Hide()
	
	f.threatVal = 0
	
	self:CreateLabels(f)

	f.SetPercent = animatePercent
	return f
end

function Omen:CreateLabels(bar)
	local skin = self.ActiveSkin
	bar.labels = {}
	local c = self.Anchor.columns
	for name, column in pairs(c) do
		local t = bar:CreateFontString(nil, nil, "GameFontNormal")
		t:SetVertexColor(1,1,1,1)
		t:SetText("-" .. name .. "-")
		bar.labels[name] = t
	end
end

function Omen:CreateColumn(name, label)
	local t = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	t:SetVertexColor(1,1,1,1)
	t.name = name
	t:SetText(label)
	self.Anchor.columns[name] = t
end

function Omen:CreateAggroGainBar(pName, name, multiplier)
	if not self.bars[pName] then
		self.bars[pName] = self:CreateBar()
		self.bars[pName].labels.name:SetText(name)
		self.bars[pName].labels.tps:SetText("")
		self.bars[pName].unitname = pName
		tinsert(self.barList,self.bars[pName])
	end
end

------------------------------------------------------
-- UI update stuff
------------------------------------------------------

function Omen:SetAnchorWidth(width)
	self.Anchor:SetWidth(width)
	self:ApplySkin()
end

function Omen:ArrangeBars()	
	if not self.Anchor:IsVisible() then return end
	local threatTarget = self.watchedUnitID

	local aggroHolderAggro = 0
	if threatTarget then	
		if Threat:UnitInMeleeRange(threatTarget) then
			aggroHolderAggro = self:UpdateAggroGainBar("_AGGRO_GAIN_", 1.1)
		else
			aggroHolderAggro = self:UpdateAggroGainBar("_AGGRO_GAIN_", 1.3)
		end
	end
	
	if not self.db.profile.ShowAggroGain and self.bars["_AGGRO_GAIN_"] then
		self.bars["_AGGRO_GAIN_"].threatVal = 0
	end
	
	table.sort(self.barList, function(a,b)
		return b.threatVal < a.threatVal
	end)
	local lastBar = nil
	local maxThreat = -1
	local barCt = 0
	local tHt = 0
	local gotPlayer = false
	local showPlayer = self.db.profile.AlwaysShowSelf
	local maxBars = self.db.profile.MaxBars
	local showKTM = self.db.profile.ShowKtmData
	
	if showPlayer then maxBars = maxBars - 1 end
	if #self.barList > 0 then	
		maxThreat = self.barList[1].threatVal
	end
	
	if aggroHolderAggro == 0 then
		aggroHolderAggro = self.barList[1] and self.barList[1].threatVal
	end
	local aggroBar = self.bars["_AGGRO_GAIN_"]
	for k,v in pairs(self.barList) do
		local isPlayer = v.unitid == "player" and showPlayer and v.threatVal > 0
		if isPlayer or (barCt < maxBars and v.threatVal > 0 and (showKTM or not v.isForeign) and (self.db.profile.Classes[v.unitclass] or v == aggroBar)) then
			v:Show()
			v:ClearAllPoints()
			
			local displayPercent = string.format("%2.1f%%", (v.threatVal / aggroHolderAggro) * 100)
			if aggroHolderAggro == 0 then
				displayPercent = "--.-%"
			end
			v:SetPercent(v.threatVal / maxThreat, displayPercent)
			if lastBar then
				v:SetPoint("TOPLEFT", lastBar, "BOTTOMLEFT", 0, 0)
				v:SetPoint("TOPRIGHT", lastBar, "BOTTOMRIGHT", 0, 0)
			else
				v:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", 5, self.barsOffset * -1)
				v:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT", -5, self.barsOffset * -1)
			end
			lastBar = v
			tHt = tHt + v:GetHeight()
			if isPlayer then
				maxBars = maxBars + 1
			end
			barCt = barCt + 1
		else
			v:Hide()
		end
	end
	self.Anchor:SetHeight(tHt + self.barsOffset + 5)
end

function Omen:Clear()
	if not self.bars then return end
	for k, v in pairs(self.bars) do
		v.threatVal = 0
		v:Hide()
	end
	self.Anchor:SetHeight(self.barsOffset + 5)
end

function Omen:UpdateAggroGainBar(pName, multiplier)
	local aggroToBeat = 0
	
	if not self.bars[pName] then return end

	if not self.watchedUnitID then return end
	for k, v in pairs(self.barList) do
		v.aggroArrow:Hide()
		if v.unitid and v.unitname == self.lastWithAggro then
			aggroToBeat = v.threatVal
			v.aggroArrow:Show()
			if Threat:UnitInMeleeRange(self.watchedUnitID) then
				multiplier = 1.1
			end
		end
	end	

	self.bars[pName].threatVal = aggroToBeat * multiplier
	local threatVal, isK = ConvertToK(self.bars[pName].threatVal)
	self.bars[pName].labels.threat:SetText(string.format("%2.1f%s", threatVal, isK))
	return aggroToBeat
end

function Omen:ShowTest()
	local pNum = GetNumPartyMembers()
	local rNum = GetNumRaidMembers()
	local addBar = self.addBar

	local uName = UnitName("player")
	if not self.bars[pName] then
		addBar(self, uName, "player")
		self.bars[uName].labels.name:SetText(uName)
	end

	uName = UnitName("pet")
	if uName then
		addBar(self, uName, "pet")
		self.bars[uName].labels.name:SetText(uName)
	end

	for i = 1, pNum do
		local pName = UnitName("party" .. i)
		if pName and not self.bars[pName] then
			addBar(self, pName, "party" .. i)
			self.bars[pName].labels.name:SetText(pName)
		end
		local pName = UnitName("party" .. i .. "pet")
		if pName and not self.bars[pName] then
			addBar(self, pName, "party" .. i .. "pet")
			self.bars[pName].labels.name:SetText(pName)
		end
	end
	
	for i = 1, rNum do
		local pName = UnitName("raid" .. i)
		if pName and not self.bars[pName] then
			addBar(self, pName, "raid" .. i)
			self.bars[pName].labels.name:SetText(pName)
		end
		local pName = UnitName("raid" .. i .. "pet")
		if pName and not self.bars[pName] then
			addBar(self, pName, "raid" .. i .. "pet")
			self.bars[pName].labels.name:SetText(pName)
		end
	end	
	
	for i = 1, (self.db.profile.MaxBars - #self.barList) do
		local pName = "* Test-" .. i
		addBar(self, pName, "player")
		self.bars[pName].labels.name:SetText(pName)
	end
		
	table.sort(self.barList, function(a,b)
		return a.unitname < b.unitname
	end)

	local lastVal = 7500
	for _, v in ipairs(self.barList) do
		lastVal = lastVal + 2500
		v.threatVal = lastVal
		v.tex:SetWidth(0)
		v.labels.threat:SetText(v.threatVal)
	end
	local v = self.bars["_AGGRO_GAIN_"]
	v.threatVal = lastVal + 2500
	v.threatVal = lastVal + 2500
	v.tex:SetWidth(0)
	v.labels.threat:SetText(v.threatVal)
	
	for k,v in ipairs(Omen.ThreatBars) do
		v:Appear()
	end
	self:ApplySkin()
end

function Omen:SetAnchors()
	local t = self.db.profile.GrowUp
	if t then
		local x, y = self.Anchor:GetLeft(), self.Anchor:GetBottom()
		Omen.Anchor:ClearAllPoints()
		Omen.Anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x, y)
	else
		local x, y = self.Anchor:GetLeft(), self.Anchor:GetTop()
		Omen.Anchor:ClearAllPoints()
		Omen.Anchor:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
	end
end
